/*****************************************************************************
 * rc v3.00 Quake 3 Arena remote console
 * Brad Bales (brad.bales@gmail.com) Copyright (c) 2002-2005
 * Created: Fri Feb  1 14:04:56 CST 2002
 * 
 * This program communicates with a server running the Quake III Arena engine,
 * although I believe it will work with the older Quake engines as well.
 * It will pass a remote console command to the server, and also gather
 * the response from the server and display it to the user.
 *
 * The user can specify command(s) on the command line, or can leave off the
 * commands and go into an interactive mode where the commands can be entered.
 *
 * Usage: rc host[:port] password [command1 command2 ...]
 *
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 *
 * In other words, do what you want with this :)
 *****************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/*****************************************************************************
 * Prototypes.
 *****************************************************************************/

static int send_command(int sd, unsigned long ipaddr, int port, char *password,
						char *command);
static void print_buffer(unsigned char *buffer, int bytes);


int main(int argc, char **argv)
{
	char *host = "localhost";
	int port = 27960;
	char *password = "";
	char *ptr;
	char line[1024];
	int sd;
	int index;
	int retcode;
	unsigned long ipaddr;
	struct hostent *hostnm;

	/*************************************************************************
	 * Command line.
	 *************************************************************************/

	if( argc < 3 ) {
		puts("Usage: rc host[:port] password [command1 command2 ...]");
		return(1);
	}

	host = argv[1];
	password = argv[2];

	/*************************************************************************
	 * Parse the port out of the hostname (if present).
	 *************************************************************************/

	if( (ptr = strchr(host, ':')) != NULL ) {

		/* Null term the hostname at this point */
		*ptr++ = '\0';

		/* Parse out the port */
		port = atoi(ptr);
	}

	/*************************************************************************
	 * Resolve the hostname if necessary, otherwise translate the dotted IP
	 * address to the binary format.
	 *************************************************************************/

	if( isdigit(*host) ) {
		ipaddr = inet_addr(host);
		if( ipaddr == -1L ) {	/* 0xffffff */
			printf("Invalid IP address.\n");
			return(1);
		}
	} else {
		hostnm = gethostbyname(host);
		if( hostnm == NULL ) {
			printf("Unknown or invalid hostname.\n");
			return(1);
		}
		ipaddr = *((unsigned long *)hostnm->h_addr);
	}

	/*************************************************************************
	 * Create the socket.
	 *************************************************************************/

	if( (sd = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ) {
		perror("Couldn't open UDP socket");
		return(1);
	}

	/*************************************************************************
	 * If no commands were issued on the command line, go into interactive
	 * mode.  Otherwise, execute all commands issued on the command line.
	 *************************************************************************/

	retcode = 0;
	if( argc == 3 ) {
		printf("Enter commands.\n\n");
		while( fgets(line, sizeof(line), stdin) != NULL ) {
			if( (ptr = strchr(line, '\r')) != NULL ) *ptr = '\0'; 
			if( (ptr = strchr(line, '\n')) != NULL ) *ptr = '\0'; 
			if( send_command(sd, ipaddr, port, password, line) ) {
				retcode = 1;
				break;
			}
		}
	} else {
		for( index = 3; index < argc; index++ ) {
			if( send_command(sd, ipaddr, port, password, argv[index]) ) {
				retcode = 1;
				break;
			}
		}
	}

	close(sd);
	return(retcode);
}

/*****************************************************************************
 * send_command():
 *
 * Given a command, send it to the Quake III server and wait for a response.
 * Once the response is received, display it on stdout.
 *
 * Returns 0 if successful; anything else if there is a problem.
 *****************************************************************************/

static int send_command(int sd, unsigned long ipaddr, int port, char *password,
						char *command)
{
	char temp_str[64];
	int len;
	int bytes;
	int got;
	socklen_t ssize;
	struct sockaddr_in to;
	struct sockaddr_in from;
	unsigned char buffer[32*1024];

	/*************************************************************************
	 * First, fill in the header information.
	 *************************************************************************/

	len = 0;
	buffer[len++] = 0xff;
	buffer[len++] = 0xff;
	buffer[len++] = 0xff;
	buffer[len++] = 0xff;
	buffer[len++] = 'r';
	buffer[len++] = 'c';
	buffer[len++] = 'o';
	buffer[len++] = 'n';
	buffer[len++] = 0x20;	/* Field separator */

	/*************************************************************************
	 * Next comes the password.
	 *************************************************************************/

	memcpy(&buffer[len], password, strlen(password));
	len += strlen(password);

	buffer[len++] = 0x20;	/* Field separator */

	/*************************************************************************
	 * Next comes the command.
	 *************************************************************************/

	memcpy(&buffer[len], command, strlen(command));
	len += strlen(command);

	memset(&to, 0, sizeof(to));
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = ipaddr; 
	to.sin_port = htons(port);

#if 0
	printf("Sending UDP packet...\n");
#endif
	bytes = sendto(sd, buffer, len, 0, (struct sockaddr *)&to, sizeof(to));
	if( bytes != len ) {
		sprintf(temp_str, "Could not send UDP packet (tried=%d, rc=%d)", len,
				bytes);
		perror(temp_str);
		return(1);
	}

	/*************************************************************************
	 * Now receive UDP packet(s) from the server.  My buffer is pretty big so
 	 * the OS will be able to concatenate multiple packets into one read, but 
	 * we can't depend upon all of the packets arriving before this recvfrom()
	 * call.  You can't even depend upon the OS to concatenate.  Therefore, if
	 * what we read isn't terminated with 0x0a, we'll read some more.  This
	 * isn't perfect; if the server sends an 0x0a terminated string as the
	 * very last thing in a packet we're going to interpret that as "there is
	 * no more" - even if there is indeed more to read.  Well, that's too bad.
	 *
	 * Please read the above statements about concatenation - it is VERY
	 * important information that ALMOST ALL SOCKET PROGRAMMERS ignore.  Lesson:
	 * unless you KNOW FOR SURE that your data is going to fit in the least
	 * common denominator MTU on your network, you have to use this kind of
	 * algorithm.  It's also better to encode an indicator in your packet so
	 * this side of the pond can determine if there's more.  Lesson ends.
	 *************************************************************************/

	bytes = 0;
	do {
#if 0
		printf("Reading %sUDP packet(s)...\n", (bytes) ? "more " : "");
#endif
		memset(&from, 0, sizeof(from));
		from.sin_family = AF_INET;
		from.sin_addr.s_addr = INADDR_ANY;
		from.sin_port = htons(port);
		ssize = sizeof(from);

		got = recvfrom(sd, &buffer[bytes], sizeof(buffer) - bytes, 0,
				(struct sockaddr *)&from, &ssize);
		if( got <= 0 ) {
			sprintf(temp_str, "Error receiving UDP packet (tried=%d, rc=%d)",
					sizeof(buffer), got);
			perror(temp_str);
			return(1);
		}

		bytes += got;

	} while( bytes < sizeof(buffer)
			 && (bytes < 5 || buffer[bytes-1] != 0x0a) );

#if 0
	printf("Bytes read = %d\n", bytes);
#endif

	/*************************************************************************
	 * Print out the contents of the buffer in a human-readable format.
	 *************************************************************************/

	print_buffer(buffer, bytes);

	return(0);
}

/*****************************************************************************
 * print_buffer():
 *
 * Print out the contents of the buffer in a human-readable format.
 *****************************************************************************/

static void print_buffer(unsigned char *buffer, int bytes)
{
	int index = 0;

	/*************************************************************************
	 * Skip past the header portion; it is always 4 bytes and is 0xff 0xff
	 * 0xff 0xff.  Go ahead and make sure it is valid.
	 *************************************************************************/

	if( buffer[0] != 0xff || buffer[1] != 0xff || buffer[2] != 0xff
	|| buffer[3] != 0xff ) {
		fprintf(stderr, "Invalid buffer header.  Sorry.\n");
		return;
	}

	index += 4;

	/*************************************************************************
	 * Now print all strings in the packet.  Each string is separated by 0x0a
	 * characters.
	 *************************************************************************/

	for( ; index < bytes; index++ ) {
		if( buffer[index] == 0x0a ) {
			putchar('\n');
		} else {
			putchar(buffer[index]);
		}
	}
	putchar('\n');
}
